/* vim: set ai et sw=4 ts=4: */
#define TEST

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <gst/gst.h>

/* to parse drag data */
#include <libgnomevfs/gnome-vfs.h>

#include "applet.h"

GramophoneApplet *gramophone;

static void gramophone_new (void);

/* ************************************************************************
 * TODO:    applet popup menuitems
************************************************************************ */
/* menu callbacks */
static void display_properties_dialog (BonoboUIComponent *uic,
                                       PanelApplet *applet,
                                       const gchar *verbname);
static void display_help_dialog (BonoboUIComponent *uic,
                                 PanelApplet *applet,
                                 const gchar *verbname);
static void display_about_dialog (BonoboUIComponent *uic,
                                  PanelApplet *applet,
                                  const gchar *verbname);

static gboolean gramophone_factory (PanelApplet *applet,
                                    const gchar *iid,
                                    gpointer data);
static gboolean gramophone_build (PanelApplet *applet);

/* applet popup menu */
static const BonoboUIVerb menu_verbs[] = {
    BONOBO_UI_UNSAFE_VERB ("GramophoneProperties", display_properties_dialog),
    BONOBO_UI_UNSAFE_VERB ("GramophoneHelp", display_help_dialog),
    BONOBO_UI_UNSAFE_VERB ("GramophoneAbout", display_about_dialog),
    BONOBO_UI_VERB_END
};

#ifdef TEST
static gboolean on_quit (GtkWidget *widget, GdkEvent *event, gpointer data)
{
    gtk_main_quit();
    return TRUE;
}

static void on_play_clicked (GtkWidget *widget, gpointer data)
{
    Engine *engine = gramophone->engine;
    engine_play (engine);
    gramophone->state = GRAMOPHONE_PLAYING;
}

static void on_pause_clicked (GtkWidget *widget, gpointer data)
{
    Engine *engine = gramophone->engine;
    engine_pause (engine);
    gramophone->state = GRAMOPHONE_PAUSED;
}

static void on_stop_clicked (GtkWidget *widget, gpointer data)
{
    Engine *engine = gramophone->engine;
    Playlist *plist = gramophone->playlist;

    engine_pause (engine);
    engine_load_uri (engine, playlist_current (plist));
    gramophone->state = GRAMOPHONE_STOPED;
}

static void on_prev_clicked (GtkWidget *widget, gpointer data)
{
    Engine *engine = gramophone->engine;
    Playlist *plist = gramophone->playlist;

    engine_pause (engine);
    engine_load_uri (engine, playlist_prev (plist));

    if (gramophone->state == GRAMOPHONE_PLAYING)
        engine_play (engine);
}

static void on_next_clicked (GtkWidget *widget, gpointer data)
{
    Engine *engine = gramophone->engine;
    Playlist *plist = gramophone->playlist;

    engine_pause (engine);
    engine_load_uri (engine, playlist_next (plist));

    if (gramophone->state == GRAMOPHONE_PLAYING)
        engine_play (engine);
}

#endif

int
main (int argc, char *argv[])
{
    bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);

    /* note: gst_init() do g_thread_init() internally. */
    gst_init (&argc, &argv);

    /* FIXME: gst_init() make --help to show gst option list.
     * then gnome_program_init() does not take place on --help option list.
     * I expected to append --help option list one by one;
     * so if gramophone_applet --help, gst option list, and gnome option list.
     */

#ifdef TEST
    gtk_init (&argc, &argv);

    do {
        GtkWidget *window, *hbox, *button;

        window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
        g_signal_connect (G_OBJECT (window), "delete_event",
                          G_CALLBACK (on_quit), NULL);

        hbox = gtk_hbox_new (TRUE, 0);
        gtk_container_add (GTK_CONTAINER (window), hbox);

        button = gtk_button_new_with_label ("Prev");
        gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
        g_signal_connect (G_OBJECT (button), "clicked",
                          G_CALLBACK (on_prev_clicked), NULL);
 
        button = gtk_button_new_with_label ("Play");
        gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
        g_signal_connect (G_OBJECT (button), "clicked",
                          G_CALLBACK (on_play_clicked), NULL);
        
        button = gtk_button_new_with_label ("Stop");
        gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
        g_signal_connect (G_OBJECT (button), "clicked",
                          G_CALLBACK (on_stop_clicked), NULL);

        button = gtk_button_new_with_label ("Pause");
        gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
        g_signal_connect (G_OBJECT (button), "clicked",
                          G_CALLBACK (on_pause_clicked), NULL);

        button = gtk_button_new_with_label ("Next");
        gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
        g_signal_connect (G_OBJECT (button), "clicked",
                          G_CALLBACK (on_next_clicked), NULL);

        gtk_widget_show_all (window);

        gramophone_new ();
    } while (0);

    gtk_main ();

    return 0;
#else
    gnome_program_init (GETTEXT_PACKAGE, VERSION,
                        LIBGNOMEUI_MODULE,
                        argc, argv,
                        GNOME_PARAM_NONE);

    return panel_applet_factory_main ("OAFIID:GNOME_GramophoneApplet_Factory",
                                      PANEL_TYPE_APPLET,
                                      gramophone_factory, NULL);
#endif
}

static gboolean
gramophone_factory (PanelApplet *applet,
                    const gchar *iid,
                    gpointer data)
{
    gboolean retval = FALSE;
    if (strcmp (iid, "OAFIID:GNOME_GramophoneApplet") == 0)
        retval = gramophone_build (applet);
    return retval;
}

static void
gramophone_new (void)
{
    gramophone = g_new0 (GramophoneApplet, 1);
    gramophone->tips = gtk_tooltips_new();

    /* playlist */
    gramophone->playlist = playlist_new ();

    /* engine */
    gramophone->engine = engine_new ();

    /* load first song */
    engine_load_uri (gramophone->engine,
                     playlist_current (gramophone->playlist));
}

static gboolean
gramophone_build (PanelApplet *applet)
{
    GtkWidget *button, *icon;
    guint applet_size;

    applet_size = panel_applet_get_size (applet);

    gramophone_new ();
    gramophone->applet = applet;

    /* widget create */
    button = gtk_toggle_button_new();
    gtk_tooltips_set_tip (gramophone->tips, button,
                          _("playlist"), "playlist button");

    /* TODO: change icon size by button size */
    icon = gtk_image_new_from_stock (GTK_STOCK_INDEX, GTK_ICON_SIZE_MENU);
    gtk_container_add (GTK_CONTAINER (button), icon);

    /* FIXME: call "change_size" signal itself, to resize button */
    gtk_widget_set_size_request (button, applet_size, applet_size);
    g_signal_connect (G_OBJECT (applet), "change_size",
                      G_CALLBACK (applet_size_changed), button);

    /* make playlist show/hide by toggle state of button. */
    g_signal_connect (G_OBJECT (button), "toggled",
                      G_CALLBACK (applet_button_toggle_playlist), gramophone);

    /* make button untoggled if playlist hide. */
    /*
    g_signal_connect (G_OBJECT (gramophone->playlist->view), "hide",
                      G_CALLBACK (applet_playlist_button_set_state), button);
     */

    /* PANEL BUG: button do not pass "popup-menu" signal to applet */
    g_signal_connect (G_OBJECT (button), "button_press_event",
                      G_CALLBACK (applet_button_press_hack), gramophone->applet);

    gtk_container_add (GTK_CONTAINER(applet), button);

    panel_applet_setup_menu_from_file (PANEL_APPLET (applet),
                                       NULL, "GNOME_GramophoneApplet.xml",
                                       NULL,
                                       menu_verbs, NULL);

    g_signal_connect (G_OBJECT (applet), "destroy",
                      G_CALLBACK (gramophone_destroy), gramophone);

    gtk_widget_show_all (GTK_WIDGET (applet));

    return TRUE;
}

void
gramophone_destroy (PanelApplet *applet,
                    GramophoneApplet *gramophone)
{
    g_print ("finalizing the engine...\n");

    /* turn off the engine */
    gst_element_set_state (GST_ELEMENT (gramophone->engine),
                           GST_STATE_NULL);
    //gst_object_unref (GST_OBJECT (gramophone->engine));

    g_print ("clean up the widgets...\n");
    /* or applet remains.. and terminal not freeed. */

    /* TODO: what to do? */
    //gtk_tree_row_reference_free (gramophone->row);
}

void
applet_size_changed (PanelApplet *applet,
                     guint size,
                     GtkWidget *widget)
{
    gtk_widget_set_size_request (widget, size, size);
}

gboolean
applet_button_press_hack (GtkWidget *w,
                          GdkEventButton *event,
                          gpointer data)
{
    PanelApplet *applet = PANEL_APPLET (data);

    if (event->button == 3) {
        GTK_WIDGET_CLASS (PANEL_APPLET_GET_CLASS (applet))->
            button_press_event (data, event);
        return TRUE;
    }
    return FALSE;
}

/* help to workaround infinite loop : Marcelo R Leitner <mrl@netbank.com.br> */
void
applet_button_toggle_playlist (GtkToggleButton *button,
                               GramophoneApplet *gramophone)
{
    if (gtk_toggle_button_get_active (button)) {
        gtk_widget_show (gramophone->playlist->view);
        gtk_tooltips_set_tip (gramophone->tips, GTK_WIDGET (button),
                              _("playlist - visible"),
                              "playlist is visible now.");
    } else {
        gtk_widget_hide (gramophone->playlist->view);
        gtk_tooltips_set_tip (gramophone->tips, GTK_WIDGET (button),
                              _("playlist"),
                              "playlist is hidden now.");
    }
}

void
display_properties_dialog (BonoboUIComponent *uic,
                           PanelApplet *applet,
                           const gchar *verbname
) {
}
void
display_help_dialog (BonoboUIComponent *uic,
                     PanelApplet *applet,
                     const gchar *verbname
) {
}
void
display_about_dialog (BonoboUIComponent *uic,
                      PanelApplet *applet,
                      const gchar *verbname
) {
}
