/* vim: set ai et sw=4 ts=4: */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <gst/gst.h>

#include <libgnomevfs/gnome-vfs-ops.h>
#include <libgnomevfs/gnome-vfs-uri.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include "applet.h"

static void on_add_clicked (GtkButton *button, gpointer data);

static gboolean     playlist_load_from_file (Playlist *plist);
static GtkWidget*   playlist_create_widget (Playlist *plist);
static GtkWidget*   playlist_create_view (Playlist *plist);

/* callbacks */
void playlist_row_activated (GtkTreeView *view,
                             GtkTreePath *path,
                             GtkTreeViewColumn *column,
                             Playlist *plist);

static void
on_f_dialog_ok_clicked (GtkFileSelection *file_selector, gpointer data)
{
	char *uri, *file;
	
	file = gtk_file_selection_get_filename (GTK_FILE_SELECTION (data));
	g_print ("%s\n", file);
}

static void
on_add_clicked (GtkButton *button, gpointer data)
{
	GtkWidget *f_dialog;
	
	f_dialog = gtk_file_selection_new ("Please select a file for playing");
	g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (f_dialog)->ok_button),
			  "clicked",
			  G_CALLBACK (on_f_dialog_ok_clicked),
			  f_dialog);

	g_siganl_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (f_dialog)->ok_button),
				  "clicked",
				  G_CALLBACK (gtk_widget_destroy),
				  NULL);
	g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (f_dialog)->cancel_button),
			  "clicked",
			  G_CALLBACK (gtk_widget_destroy),
			  NULL);

	gtk_widget_show (f_dialog);
}

Playlist*
playlist_new (void)
{
    Playlist *plist;        /* playlist */
    GtkTreePath *path;

    plist = g_new0 (Playlist, 1);
    plist->window = NULL;
    plist->view = NULL;
    plist->model = NULL;

    /* make widget */
    plist->window = playlist_create_widget (plist);

    /* loading list file */
    playlist_load_from_file (plist);

    /* TODO: set this from saved value */
    path = gtk_tree_path_new_from_string ("0");
    plist->current = gtk_tree_row_reference_new (plist->model, path);

    return plist;
}

GtkWidget*
playlist_create_widget (Playlist *plist)
{
    GtkWidget *window, *view, *vbox, *scrolled;
    GtkWidget *hbox, *add;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    plist->window = window;
    gtk_window_set_resizable (GTK_WINDOW (window), TRUE);

    view = playlist_create_view (plist);
    plist->view = view;
    g_signal_connect (G_OBJECT (view), "row_activated",
                      G_CALLBACK (playlist_row_activated), plist);

    scrolled = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled),
                                         GTK_SHADOW_ETCHED_IN);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled),
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_ALWAYS);

    gtk_container_add (GTK_CONTAINER (scrolled), view);

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), vbox);
    gtk_box_pack_start (GTK_BOX (vbox), scrolled, TRUE, TRUE, 0);

    hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
    
    add = gtk_button_new_with_label ("Add");
    gtk_box_pack_start (GTK_BOX (hbox), add, FALSE, FALSE, 0);
    g_signal_connect (G_OBJECT (add), "clicked",
		      G_CALLBACK (on_add_clicked), NULL);

    gtk_widget_show_all (window);

    return window;
}

gboolean
playlist_load_from_file (Playlist *plist)
{
    FILE *list_file;
    char buf[1024];
    gchar* filename;
    gchar* list_file_name;
    GtkTreeIter iter;
    //GtkTreeIter child_iter;
    GtkTreeStore *store = GTK_TREE_STORE (plist->model);

    list_file_name = g_strconcat (g_get_home_dir(),
                                  "/.gramophone/playlist",
                                  NULL);
    list_file = fopen (list_file_name, "r");
    g_assert (list_file != NULL);

    while (fgets (buf, sizeof (buf), list_file) != NULL) {
        g_strchomp (buf);
        filename = g_filename_from_uri (buf, NULL, NULL);

        gtk_tree_store_append (store, &iter, NULL);
        gtk_tree_store_set (store, &iter,
                            PLAYLIST_URI_COLUMN, buf,
                            PLAYLIST_FILENAME_COLUMN, filename,
                            -1);
        g_free (filename);
    }

    fclose (list_file);
    g_free (list_file_name);

    return TRUE;
}

GtkWidget*
playlist_create_view (Playlist *plist)
{
    GtkTreeStore *store;
    GtkCellRenderer *cell_renderer;
    GtkTreeViewColumn *column;
    GtkWidget *view;

    store = gtk_tree_store_new (N_COLS,
                                G_TYPE_STRING,
                                G_TYPE_STRING);
    plist->model = GTK_TREE_MODEL (store);

    cell_renderer = gtk_cell_renderer_text_new ();

    column = gtk_tree_view_column_new_with_attributes (_("filename"),
                                                       cell_renderer,
                                                       "text",
                                                       PLAYLIST_FILENAME_COLUMN,
                                                       NULL);

    view = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
    gtk_tree_view_append_column (GTK_TREE_VIEW (view), column);

    return view;
}

void
playlist_free (Playlist *plist)
{
    gtk_tree_row_reference_free (plist->current);
    g_free (plist);
}

const gchar*
playlist_current (Playlist *plist)
{
    gchar *text_uri;
    GtkTreePath *path;
    GtkTreeIter iter;

    path = gtk_tree_row_reference_get_path (plist->current);
    gtk_tree_model_get_iter (plist->model, &iter, path);
    gtk_tree_model_get (plist->model, &iter,
                        PLAYLIST_URI_COLUMN, &text_uri,
                        -1);
    gtk_tree_path_free (path);

    return text_uri;
}

const gchar*
playlist_prev (Playlist *plist)
{
    gboolean valid;
    gchar *text_uri;
    GtkTreePath *path;
    GtkTreeIter iter;

    path = gtk_tree_row_reference_get_path (plist->current);
    valid = gtk_tree_path_prev (path);
    if (valid) {
        valid = gtk_tree_model_get_iter (plist->model, &iter, path);
        if (valid) {
            gtk_tree_model_get (plist->model, &iter,
                                PLAYLIST_URI_COLUMN, &text_uri,
                                -1);
            g_print ("URI: %s\n", text_uri);

            /* saving row ref */
            gtk_tree_row_reference_free (plist->current);
            plist->current = gtk_tree_row_reference_new (plist->model, path);

            return text_uri;
        }
    }

    return NULL;
}

const gchar*
playlist_next (Playlist *plist)
{
    gboolean valid;
    gchar *text_uri;
    GtkTreePath *path;
    GtkTreeIter iter;

    path = gtk_tree_row_reference_get_path (plist->current);
    gtk_tree_path_next (path);
    valid = gtk_tree_model_get_iter (plist->model, &iter, path);
    if (valid) {
        gtk_tree_model_get (plist->model, &iter,
                            PLAYLIST_URI_COLUMN, &text_uri,
                            -1);
        g_print ("URI: %s\n", text_uri);

        /* saving row ref */
        gtk_tree_row_reference_free (plist->current);
        plist->current = gtk_tree_row_reference_new (plist->model, path);

        return text_uri;
    }

    return NULL;
}

void
playlist_row_activated (GtkTreeView *view,
                        GtkTreePath *path,
                        GtkTreeViewColumn *column,
                        Playlist *plist)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    gchar *text_uri;
    GnomeVFSURI *uri;

    /* get file location(uri) from GtkTreeStore */
    model = gtk_tree_view_get_model (view);

    gtk_tree_model_get_iter (model, &iter, path);
    gtk_tree_model_get (model, &iter,
                        PLAYLIST_URI_COLUMN, &text_uri,
                        -1);

    if (text_uri == NULL) {
        // NEED: error handling
        g_print("URI is not valid: NULL\n");
        return;
    }

    /* save current position */
    gtk_tree_row_reference_free (plist->current);
    plist->current = gtk_tree_row_reference_new (model, path);
    g_print ("saved row to play now.\n");

    /* Does we really need URI check ?
     * Just for file existence check, we make new object
     * I think this is too expensive */
    uri = gnome_vfs_uri_new (text_uri);
    if (gnome_vfs_uri_exists (uri)) {
        engine_pause (gramophone->engine);
        engine_load_uri (gramophone->engine, text_uri);
        engine_play (gramophone->engine);
        gtk_tree_view_scroll_to_cell (view, path, column, TRUE, 0.5, 0.0);
    } else {
        /* TODO: uri is broken. */
        g_print ("there is no fuel there.\n");

        /* turn off the engine? */
        /* remove the row? */
    }
    gnome_vfs_uri_unref (uri);
}




// old funcs

#if 0

GtkWidget*
playlist_list_new (void)
{
    GtkTreeModel *model;
    GtkCellRenderer *cell_renderer;
    GtkTreeViewColumn *column;
    GtkWidget *view;

    /* FIXME: more accurate types */
    const GtkTargetEntry dnd[] = {
        /*
        { "text/plain", 0, WEB },
        */
        { "x-url/http", 0, WEB },
        { "_NETSCAPE_URL", 0, WEB }
    };

    model = GTK_TREE_MODEL (gtk_list_store_new (N_COLS,
                                                G_TYPE_STRING,
                                                G_TYPE_STRING));

    cell_renderer = gtk_cell_renderer_text_new ();

    column = gtk_tree_view_column_new_with_attributes (_("filename"),
                                                       cell_renderer,
                                                       "text",
                                                       PLAYLIST_FILENAME_COLUMN,
                                                       NULL); /* arg end */
    gtk_tree_view_column_set_clickable (column, TRUE);

    view = gtk_tree_view_new_with_model (model);
    gtk_tree_view_append_column (GTK_TREE_VIEW (view), column);

    gtk_drag_dest_set (view, GTK_DEST_DEFAULT_ALL,
                       dnd, G_N_ELEMENTS(dnd), GDK_ACTION_COPY);
    g_signal_connect (G_OBJECT (view), "drag_data_received",
                      G_CALLBACK (playlist_get_drop), model);

    /* TODO: tree model/view built-in drag-n-drop.
    gtk_tree_view_enable_model_drag_dest (GTK_TREE_VIEW (view),
                                          dnd, G_N_ELEMENTS (dnd),
                                          GDK_ACTION_COPY);
    */

    return view;
}

GtkWidget*
playlist_toolbar_new (GtkWidget *view,
                      GramophoneApplet *gramophone)
{
    GtkWidget *toolbar, *item;
    
    /* TODO: give model directly to button's callback, AMAP.
    GtkTreeModel *model;
    model = gtk_tree_view_get_model (GTK_TREE_VIEW (view));
    */

    /* TODO: bonoboui */
    toolbar = gtk_toolbar_new();
    gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_ICONS);

    /* TODO:
     * give pointer of GtkTreeView directly, not gramophone big thing,
     * to make easy to play with.
     */
    item = gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                                     GTK_STOCK_GO_FORWARD,
                                     _("play button tooltip"),
                                     _("play button tooltip private"),
                                     G_CALLBACK(playlist_play_button_clicked),
                                     "callback data - play button",
                                     -1 /* is append, 0 is prepend */);
    item = gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                                     GTK_STOCK_STOP,
                                     _("stop button tooltip"),
                                     _("stop button tooltip private"),
                                     G_CALLBACK(playlist_button_stop),
                                     gramophone,
                                     -1);
    item = gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                                     GTK_STOCK_GO_UP,
                                     _("prev button tooltip"),
                                     _("prev button tooltip private"),
                                     G_CALLBACK(playlist_button_prev),
                                     gramophone,
                                     -1);
    item = gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                                     GTK_STOCK_GO_DOWN,
                                     _("next button tooltip"),
                                     _("next button tooltip private"),
                                     G_CALLBACK(playlist_button_next),
                                     gramophone,
                                     -1);

    gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

    /* trash icon, to remove: by dnd or click. */
    item = gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                GTK_STOCK_DELETE,
                _("remove button tooltip"),
                _("remove button tooltip private"),
                G_CALLBACK(playlist_button_remove),
                gramophone,
                -1);
    /* TODO: remove button accept drop from list.
    g_signal_connect (G_OBJECT (item), "drag_data_get",
                      G_CALLBACK (playlist_remove_button_get_drop),
                      "callback data. GtkListStore is core.");
    */

    gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

    item = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
                                       GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL,
                                       _("Loop"),
                                       _("loop toggle button tooltip"),
                                       _("loop toggle button tooltip private"),
                                       /* FIXME: flexible size */
                                       gtk_image_new_from_stock(
                                           GTK_STOCK_REFRESH,
                                           GTK_ICON_SIZE_LARGE_TOOLBAR
                                       ),
                                       G_CALLBACK(playlist_button_loop),
                                       gramophone);

    return toolbar;
}

void
playlist_play_button_clicked (GtkWidget *button,
                              gpointer callback_data)
{
    g_print ("play button clicked.\n");
}

void
playlist_button_stop (GtkWidget *button,
                      GramophoneApplet *gramophone)
{
    g_print ("you told me stop.\n");

    g_print ("remove row reference.\n"); /* or not? */
    gramophone->row = NULL;

    g_print ("turn off the engine.\n");
    gst_element_set_state (GST_ELEMENT (gramophone->engine), GST_STATE_NULL);
}

void
playlist_button_prev (GtkWidget *button,
                      GramophoneApplet *gramophone)
{
    GtkWidget *bin, *view;
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;

    bin = playlist_get_child (gramophone->playlist->view, GTK_TYPE_SCROLLED_WINDOW);
    g_return_if_fail (GTK_IS_BIN(bin));

    view = gtk_bin_get_child (GTK_BIN(bin));
    g_return_if_fail (GTK_IS_TREE_VIEW (view));

    model = gtk_tree_view_get_model (GTK_TREE_VIEW (view));
    path = playlist_go_prev_row (gramophone->row, model, &iter);

    gtk_tree_view_row_activated (GTK_TREE_VIEW (view), path,
                                 gtk_tree_view_get_column (GTK_TREE_VIEW (view),
                                                           PLAYLIST_FILENAME_COLUMN));
    
    gtk_tree_path_free (path);
}

void playlist_button_next(
        GtkWidget *button,
        GramophoneApplet *gramophone
) {
    GtkWidget *bin, *view;
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;

    bin = playlist_get_child (gramophone->playlist->view,
                GTK_TYPE_SCROLLED_WINDOW);
    g_return_if_fail (GTK_IS_BIN(bin));

    view = gtk_bin_get_child (GTK_BIN(bin));
    g_return_if_fail (GTK_IS_TREE_VIEW (view));

    model = gtk_tree_view_get_model (GTK_TREE_VIEW (view));
    path = playlist_go_next_row (gramophone->row, model, &iter);

    gtk_tree_view_row_activated (GTK_TREE_VIEW (view), path,
                                 gtk_tree_view_get_column (GTK_TREE_VIEW (view),
                                                           PLAYLIST_FILENAME_COLUMN));
    
    gtk_tree_path_free (path);
}

void
playlist_button_remove (GtkWidget *button,
                        GramophoneApplet *gramophone)
{
    GtkWidget *bin, *view;
    GtkTreeSelection *sel;

    bin = playlist_get_child (gramophone->playlist->view, GTK_TYPE_SCROLLED_WINDOW);
    g_return_if_fail (GTK_IS_BIN(bin));

    view = gtk_bin_get_child (GTK_BIN(bin));
    g_return_if_fail (GTK_IS_TREE_VIEW (view));

    sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
    g_print ("I got selection of treeview.\n");

    g_print ("remove them.\n");
    gtk_tree_selection_selected_foreach (sel, playlist_remove_row, NULL);
}

/* call by gtk_tree_selection_selected_foreach() */
void
playlist_remove_row (GtkTreeModel *model,
                     GtkTreePath *path,
                     GtkTreeIter *iter,
                     gpointer callback_data)
{
    gtk_list_store_remove (GTK_LIST_STORE (model), iter);
}

void
playlist_remove_button_get_drop ()
{
}

void
playlist_button_loop (GtkToggleButton *button,
                      GramophoneApplet *gramophone)
{
    g_object_set_data (G_OBJECT(gramophone->playlist), "loop_boolean",
                       GSIZE_TO_POINTER
                           (gtk_toggle_button_get_active (button)));
}

void
applet_playlist_button_set_state (GtkWidget *playlist,
                                  GtkToggleButton *button)
{
    gtk_toggle_button_set_active (button, FALSE);
}

void
playlist_get_drop (GtkWidget *widget,
                   GdkDragContext *context,
                   gint x, gint y,
                   GtkSelectionData *selection_data,
                   guint info,
                   guint time,
                   GtkTreeModel *model)
{
    GtkTreeIter iter;
    gchar *mime_string, *raw_data, *display_data;
    GList *elem, *uris;

    /* FIXME: nautilus drag data contain only one item. is this my bug? */
    uris = gnome_vfs_uri_list_parse (selection_data->data);
    /* returns GList of GnomeVFSURI. */

    gtk_drag_finish (context, TRUE, FALSE, time);

    for (elem = uris;
         elem != NULL && elem->data != NULL;
         elem = g_list_next (elem))
    {
        GnomeVFSURI *uri = elem->data;
        elem->data = gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE);
        gnome_vfs_uri_unref (uri);

        mime_string = (gchar *)gnome_vfs_get_mime_type (elem->data);
        /* TODO: way to refer external mime list to allow, like:
            gboolean mime_is_in_list(
                    const gchar *mime_string
            ) {
                // or g_list_find() better?
                while (list) { // 'list' made by pref.
                    if (!strcmp (list->data, mime_string))
                        return TRUE;
                    list = g_list_next (list);
                }
                return FALSE;
            }
        */
        /* strcmp() return 0 if TRUE. not prefix with '!' */
        if (strcmp (mime_string, "audio/mpeg")
         && strcmp (mime_string, "audio/x-mp3")
         && strcmp (mime_string, "audio/x-riff")
         && strcmp (mime_string, "application/x-ogg")
        ) {
            GtkWidget *dialog_if_invalid_mime;
            dialog_if_invalid_mime
                = gtk_message_dialog_new (NULL,
                    GTK_DIALOG_DESTROY_WITH_PARENT,
                    GTK_MESSAGE_ERROR,
                    GTK_BUTTONS_CLOSE,
                    _("MIME type '%s' not allowed."),
                    mime_string);
            gtk_dialog_run (GTK_DIALOG (dialog_if_invalid_mime));
            gtk_widget_destroy (dialog_if_invalid_mime);
            continue;
        }
        g_free (mime_string);

        raw_data = g_strdup (elem->data);

        /* FIXME: this uri API require GnomeVFSURI type of arg.
        display_data = gnome_vfs_uri_extract_short_name (display_data);
        g_print ("basename: %s\n", display_data);
        */
        display_data = gnome_vfs_unescape_string (raw_data, "/");
        g_print ("unescape: %s\n", display_data);

        /* FIXME: return NULL
        display_data = g_locale_to_utf8 (display_data,
                        -1, NULL, NULL, NULL);
        */
        display_data = g_convert (display_data, -1, "UTF-8", "EUC-KR",
                                  NULL, NULL, NULL);
        g_print ("utf8: %s\n", display_data);

        gtk_list_store_append (GTK_LIST_STORE (model), &iter);
        gtk_list_store_set (GTK_LIST_STORE (model), &iter,
                            PLAYLIST_FILENAME_COLUMN, raw_data,
                            PLAYLIST_FILENAME_COLUMN, display_data,
                            -1);
        g_free (raw_data);
        g_free (display_data);
    }
    /* FIXME: crash */
    //gnome_vfs_uri_list_free (uris);
}

GtkWidget*
playlist_get_child (GtkWidget *dialog,
                    GtkType type)
{
    GList *list = NULL;
    GtkBoxChild *container = NULL;
    GtkWidget *child = NULL;

    g_return_val_if_fail (GNOME_IS_DIALOG (dialog), NULL);
    list = GTK_BOX (GNOME_DIALOG (dialog)->vbox)->children;
    while (list) {
        container = (GtkBoxChild *)list->data;
        if (GTK_CHECK_TYPE (container->widget, type))
            child = container->widget;
        list = g_list_next (list);
    }
    g_return_val_if_fail (GTK_IS_SCROLLED_WINDOW (child), NULL);

    return child;
}

GtkTreePath*
playlist_go_prev_row (GtkTreeRowReference *current_row,
                      GtkTreeModel *model,
                      GtkTreeIter *iter /* given by reference */
) {
    GtkTreePath *path = NULL;

    if (current_row != NULL) {
        path = gtk_tree_row_reference_get_path (current_row);
        gtk_tree_path_prev (path);
    } else
    if (gtk_tree_model_get_iter_first (model, iter)) {
        gtk_tree_model_iter_nth_child (model, iter, /* parent */ NULL,
                                       /* number of toplevel nodes, if NULL iter */
                                       gtk_tree_model_iter_n_children (model,
                                                                       NULL) - 1);
        path = gtk_tree_model_get_path (model, iter);
    } else {
        /* no row in playlist */
    }

    return path;
}

GtkTreePath *
playlist_go_next_row (GtkTreeRowReference *current_row,
                      GtkTreeModel *model,
                      GtkTreeIter *iter)
{
    GtkTreePath *path = NULL;

    if (current_row != NULL) {
        path = gtk_tree_row_reference_get_path (current_row);
        gtk_tree_model_get_iter (model, iter, path);
        if (gtk_tree_model_iter_next (model, iter)) {
            path = gtk_tree_model_get_path (model, iter);
        }
    } else
    if (gtk_tree_model_get_iter_first (model, iter)) {
        /* first row */
        path = gtk_tree_model_get_path (model, iter);
    } else {
        /* no row in playlist */
    }

    return path;
}

#endif
